<?php

namespace Drupal\Tests\migrate_sandbox\FunctionalJavascript;

use Drupal\FunctionalJavascriptTests\WebDriverTestBase;
use Drupal\node\Entity\Node;
use Drupal\node\Entity\NodeType;

/**
 * Tests migrate_sandbox.
 *
 * @group migrate_sandbox
 */
class MigrateSandboxTest extends WebDriverTestBase {

  /**
   * Modules to install.
   *
   * @var array
   */
  protected static $modules = [
    'migrate',
    'migrate_sandbox',
    'node',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * Tests the migrate_sandbox form.
   */
  public function testMigrateSandbox() {
    // Test as trusted user.
    $trusted_user = $this->drupalCreateUser(['access migrate_sandbox']);
    $this->drupalLogin($trusted_user);
    $this->drupalGet('/admin/config/development/migrate-sandbox');
    $page = $this->getSession()->getPage();
    $assert_session = $this->assertSession();

    // Test that urlencode works.
    $page->find('css', '[name="prepopulated"]')->setValue('urlencode');
    $assert_session->assertWaitOnAjaxRequest();
    $page->find('css', '[name="constants"]')->setValue('hello: howdy');
    $assert_session->pageTextNotContains("new_url: 'http://example.com/a%20url%20with%20spaces.html'");
    $page->pressButton('Save & Run');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->pageTextContains("new_url: 'http://example.com/a%20url%20with%20spaces.html'");

    // Test that config was saved correctly.
    $latest = \Drupal::configFactory()->get('migrate_sandbox.latest');
    $this->assertSame('migrate_sandbox', $latest->get('id'));
    $this->assertSame('Migrate Sandbox', $latest->get('label'));
    $this->assertSame('embedded_data', $latest->get('source.plugin'));
    $this->assertSame('http://example.com/a url with spaces.html', $latest->get('source.data_rows')[0]['my_url_with_spaces']);
    $this->assertSame('123', $latest->get('source.data_rows')[0]['migrate_sandbox_dummy_source_id']);
    $this->assertSame('howdy', $latest->get('source.constants.hello'));
    $this->assertSame('urlencode', $latest->get('process.new_url.plugin'));
    $this->assertSame('my_url_with_spaces', $latest->get('process.new_url.source'));
    $this->assertSame('migrate_sandbox_config', $latest->get('destination.plugin'));
    $this->assertSame('migrate_sandbox.processed', $latest->get('destination.config_name'));

    // That migrate map does not exist.
    $this->assertFalse(\Drupal::database()->schema()->tableExists('migrate_map_migrate_sandbox'));
    $this->assertFalse(\Drupal::database()->schema()->tableExists('migrate_message_migrate_sandbox'));

    // Test sandbox escape warnings.
    $callback_warning = $page->find('css', '[data-plugin="callback"]');
    $lookup_warning = $page->find('css', '[data-plugin="migration_lookup"]');
    $this->assertFalse($callback_warning->isVisible());
    $this->assertFalse($lookup_warning->isVisible());
    $page->find('css', '[name="process"]')->setValue('plugin: callback');
    $assert_session->waitForElementVisible('css', '[data-plugin="callback"]');
    $this->assertTrue($callback_warning->isVisible());
    $this->assertFalse($lookup_warning->isVisible());
    $page->find('css', '[name="process"]')->setValue('plugin: "migration_lookup"');
    $assert_session->waitForElementVisible('css', '[data-plugin="migration_lookup"]');
    $this->assertFalse($callback_warning->isVisible());
    $this->assertTrue($lookup_warning->isVisible());

    // Try submitting with invalid yaml.
    $page->find('css', '[name="data_rows"]')->setValue(' foo:');
    $assert_session->elementNotExists('css', 'div[aria-label="Error message"]');
    $assert_session->elementTextContains('css', '[data-drupal-selector="edit-yaml"]', "new_url: 'http://example.com/a%20url%20with%20spaces.html'");
    $page->pressButton('Save & Run');
    $assert_session->waitForElementVisible('css', 'div[aria-label="Error message"]');
    // The results text should be cleared.
    $assert_session->elementTextNotContains('css', '[data-drupal-selector="edit-yaml"]', "new_url: 'http://example.com/a%20url%20with%20spaces.html'");

    // Run log example to test for on-screen messages.
    $page->find('css', '[name="prepopulated"]')->setValue('log');
    $assert_session->assertWaitOnAjaxRequest();
    $expected_log = "'example_1' value is 'Foo'";
    $assert_session->pageTextNotContains($expected_log);
    $assert_session->elementTextNotContains('css', '[data-drupal-selector="edit-yaml"]', 'example_1: Foo');
    $page->pressButton('Save & Run');
    $assert_session->waitForElementVisible('css', 'div[aria-label="Error message"]');
    $assert_session->elementTextContains('css', 'div[aria-label="Error message"]', $expected_log);
    $assert_session->elementTextContains('css', '[data-drupal-selector="edit-yaml"]', 'example_1: Foo');
    $this->assertFalse(\Drupal::database()->schema()->tableExists('migrate_map_migrate_sandbox'));
    $this->assertFalse(\Drupal::database()->schema()->tableExists('migrate_message_migrate_sandbox'));

    // Run example that triggers migrate exception.
    $page->find('css', '[name="prepopulated"]')->setValue('explode');
    $assert_session->assertWaitOnAjaxRequest();
    // Run successfully once to populate results.
    $page->pressButton('Save & Run');
    $assert_session->assertWaitOnAjaxRequest();
    $page->find('css', '[name="data_rows"]')->setValue('');
    $expected_log = 'NULL is not a string';
    $assert_session->pageTextNotContains($expected_log);
    $assert_session->elementTextContains('css', '[data-drupal-selector="edit-yaml"]', 'example_1:');
    $page->pressButton('Save & Run');
    $assert_session->waitForElementVisible('css', 'div[aria-label="Error message"]');
    $assert_session->elementTextContains('css', 'div[aria-label="Error message"]', $expected_log);
    // Result should be cleared.
    $assert_session->elementTextNotContains('css', '[data-drupal-selector="edit-yaml"]', 'example_1:');

    // Test sandbox content entity destination.
    NodeType::create(['type' => 'page'])->save();
    \Drupal::service('plugin.manager.migrate.destination')->getDefinition('migrate_sandbox_entity:node');
    // Get a clean slate.
    $this->drupalGet('/admin/config/development/migrate-sandbox');
    $page->find('css', '[name="prepopulated"]')->setValue('null_coalesce');
    $assert_session->assertWaitOnAjaxRequest();
    $page->find('css', '[name="destination"]')->setValue('entity');
    $page->find('css', '[name="entity_type"]')->setValue('node');
    $page->find('css', '[name="entity_bundle"]')->setValue('page');
    // Expect validation error since title is not in process.
    $page->pressButton('Save & Run');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->pageTextContains('(Migrate Message)');
    $page->find('css', '[name="data_rows"]')->setValue('my_string: My title');
    $page->find('css', '[name="process"]')->setValue('title: my_string');
    $assert_session->pageTextNotContains("value: 'My title'");
    $page->pressButton('Save & Run');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->pageTextContains("value: 'My title'");

    // Confirm node was not saved.
    $nodes = Node::loadMultiple();
    $this->assertEmpty($nodes);
  }

}
